//
//  Board.c
//  fornax2_core
//
//  Created by Anders on 26.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#include <stdlib.h>
#include "board.h"
#include "bitboard.h"
#include "definitions.h"
#include "../moving/movegen.h"
#include "../tt/zobrist.h"
#include "../genericutils/require.h"
#include "../eval/eval_piecetables.h"
#include "../eval/eval_definitions.h"

inhabitant board_get_inhabitant(const Board* board, square square, bool isWhite) {
  const bitboard64* array = isWhite ? board->white.pieces : board->black.pieces;
  for (int i = 0; i < 6; ++i) {
    if (bitboard_test(array[i], square)) {
      return i;
    }
  }
  return -1;
}

void board_set_inhabitant(Board* board, bool isWhite, square square, inhabitant inhabitant) {
  BoardSide* side = isWhite ? &board->white : &board->black;
  BITBOARD_SET(side->pieces[inhabitant], square);
  board->capturemap[square] = inhabitant;
}

inhabitant board_get_captured(const Board* board, square pos) {
  const bitboard64* array = board->isWhiteMove ? board->black.pieces : board->white.pieces;
  inhabitant capture = 0;
  
  for (int i = 0; i < 6; ++i) {
    if (bitboard_test(array[i], pos)) {
      capture = i;
      assert(capture != INHABITANT_KINGS);
      break;
    }
  }
  assert((capture > 0 && bitboard_test(board->hostile->pieceMask, pos)) || (capture == 0 && !bitboard_test(board->hostile->pieceMask, pos)));
  assert( capture == board->capturemap[pos]);
  
  return capture;
}

bitboard64 board_pieces_mask(bitboard64* pieceArray) {
  bitboard64 mask = *pieceArray++;
  mask |= *pieceArray++;
  mask |= *pieceArray++;
  mask |= *pieceArray++;
  mask |= *pieceArray++;
  mask |= *pieceArray;
  return mask;
}

eval board_count_material(const BoardSide* side) {
  eval material = 0;
  
  for (int i = 0; i < 6; ++i) {
    material += BITBOARD_COUNT(side->pieces[i]) * PIECEVALUE[i];
  }
  return material;
}

eval board_count_positional(const BoardSide* side, const piecetable* table) {
  eval positional = 0;
  
  for (int i = 0; i < 6; ++i) {
    bitboard64 pieces = side->pieces[i];
    square pos;
    while ((pos = bitboard_popTrailing(&pieces)) != SQUARE_NONE) {
      positional += (*table)[i][pos];
    }
  }
  return positional;
}

eval board_count_phase(const BoardSide* side) {
  uint8_t phase = 0;
  
  for (int i = 0; i < 6; ++i) {
    phase += BITBOARD_COUNT(side->pieces[i]) * PHASEVALUE[i];
  }
  return phase;
}

void board_set_friendlies_and_hostiles(Board* board) {
  
  BoardSide* white = &board->white;
  BoardSide* black = &board->black;
  
  white->pieceMask = board_pieces_mask(white->pieces);
  white->eval_material = board_count_material(white);
  white->game_phase = board_count_phase(white);
  white->mg_piecetable = &WHITE_MG_PIECETABLE;
  white->eval_mg_positional = board_count_positional(white, white->mg_piecetable);
  white->eg_piecetable = &WHITE_EG_PIECETABLE;
  white->eval_eg_positional = board_count_positional(white, white->eg_piecetable);
  
  black->pieceMask = board_pieces_mask(black->pieces);
  black->eval_material = board_count_material(black);
  black->game_phase = board_count_phase(black);
  black->mg_piecetable = &BLACK_MG_PIECETABLE;
  black->eval_mg_positional = board_count_positional(black, black->mg_piecetable);
  black->eg_piecetable = &BLACK_EG_PIECETABLE;
  black->eval_eg_positional = board_count_positional(black, black->eg_piecetable);
  
  board->friendly = board->isWhiteMove ? &board->white : &board->black;
  board->hostile = board->isWhiteMove ? &board->black : &board->white;
}

static ttkey board_getHash(const Board* board) {
  return zobrist_get(board);
}

void board_copy(const Board* original, Board* copy) {
  
  *copy->copyState = *original->copyState;
  BoardCopiedState* state = copy->copyState;
  
  *copy = *original;
  copy->copyState = state;
  copy->friendly = copy->isWhiteMove ? &copy->white : &copy->black;
  copy->hostile = copy->isWhiteMove ? &copy->black : &copy->white;
}

void board_clear(Board* board) {
  
  BoardCopiedState emptyState = { 0 };
  *board->copyState = emptyState;
  BoardCopiedState* state = board->copyState;
  
  Board empty = { 0 };
  *board = empty;
  board->copyState = state;
  
  board->friendly = &board->black;
  board->hostile = &board->white;
}

void board_flip(const Board* board, Board* copy) {
  
  for (int i = 0; i < 6; ++i) {
    copy->white.pieces[i] = BITBOARD_FLIP(board->black.pieces[i]);
    copy->black.pieces[i] = BITBOARD_FLIP(board->white.pieces[i]);
  }
  copy->isWhiteMove = !board->isWhiteMove;
  copy->isCheck = board->isCheck;
  
  board_set_friendlies_and_hostiles(copy);
  copy->copyState->enPassant = BITBOARD_FLIP(board->copyState->enPassant);
  copy->copyState->fcq = board->copyState->hcq;
  copy->copyState->fck = board->copyState->hck;
  copy->copyState->hcq = board->copyState->fcq;
  copy->copyState->hck = board->copyState->fck;
  
  copy->copyState->hash = board_getHash(copy);
  movegen_genmaps(copy);
  movegen_genmaps_hostile_attacks(copy);
  
  copy->unchangedPawnStructure = false;

}

Board* board_allocate(void) {

  BoardCopiedState* state = malloc(sizeof (BoardCopiedState));
  Board* board = malloc(sizeof (Board));
  board->copyState = state;
  return board;;
}

void board_free(Board* board) {
  free(board->copyState);
  free(board);
}
