//
//  boardprint.c
//  fornax2_core
//
//  Created by Anders on 15/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include <string.h>
#include <ctype.h>
#include "genericutils/require.h"
#include "genericutils/iolog.h"
#include "boardprint.h"
#include "board/bitboard.h"
#include "board/square.h"
#include "board/move.h"
#include "fen/fen.h"
#include "tt/transpositions.h"
#include "moving/makemove.h"

void boardprint_printbest(move m) {
  iolog_printf("bestmove ");
  if (m) boardprint_move(m);
  else iolog_printf("none");
  iolog_printf("\n");
  iolog_flush();
}

void boardprint_printcurrmove(move m, uint8_t depth, uint8_t index) {
  iolog_printf("info depth %d currmove ", depth);
  if (m) boardprint_move(m);
  else iolog_printf("none");
  iolog_printf(" currmovenumber %d\n", index);
  iolog_flush();
}

void boardprint_printpv(const Board* b, move m, uint8_t depth) {
  if (depth == 0) return;
  if (!m) {
    iolog_printf("none");
    return;
  }
  boardprint_move(m);
  Board copy = BOARD_COPY(copy, b);
  
  BoardCopiedState nextstate;
  makemove_make(&copy, &nextstate, m);
  Bucket* bucket = transpositions_get(copy.copyState->hash);
  if (bucket->hash != copy.copyState->hash || bucket->move == 0) return;
  iolog_printf(" ");
  boardprint_printpv(&copy, bucket->move, --depth);
}

char boardprint_getSymbolOfType(int inhabitant) {
  switch (inhabitant) {
    case INHABITANT_ROOKS:
      return 'r';
    case INHABITANT_KNIGHTS:
      return 'n';
    case INHABITANT_BISHOPS:
      return 'b';
    case INHABITANT_QUEENS:
      return 'q';
    case INHABITANT_KINGS:
      return 'k';
    case INHABITANT_PAWNS:
      return 'p';
      
    default:
      require(0, "unknown piecetype: %d", inhabitant);
  }
  return '?';
}

void boardprint_square(const Square sq) {
  iolog_printf("%c%c", square_to_file(sq), square_to_uci_rank(sq));
}

void boardprint_move(move m) {
  if (m == 0) {
    iolog_printf("none");
    return;
  }
  boardprint_square(Square_unflatten(MOVE_ORIGIN(m)));
  boardprint_square(Square_unflatten(MOVE_DESTINATION(m)));
  if (MOVE_IS_PROMOTION(m)) {
    iolog_printf("%c", boardprint_getSymbolOfType(MOVE_DESTINATION_PIECE(m)));
  }
}

void boardprint_debugmove(move m) {
  inhabitant originp = MOVE_ORIGIN_PIECE(m);
  inhabitant destp = MOVE_DESTINATION_PIECE(m);
  inhabitant capturep = MOVE_CAPTURE(m);
  
  boardprint_move(m);
  iolog_printf(" %c",
               boardprint_getSymbolOfType(originp));
  if (originp != destp) {
    iolog_printf("=%c", boardprint_getSymbolOfType(destp));
  }
  iolog_printf(" [value = %d]", (int8_t) MOVE_VALUE(m));
  if (capturep) {
    iolog_printf(" [capture = %c]", boardprint_getSymbolOfType(capturep));
  }
  iolog_printf("\n");
}

static char getSymbol(const Board b, int x, int y) {
  for (int i = 0; i < 6; ++i) {
    if (bitboard_test(b.white.pieces[i], SQUARE_FLATTEN(x, y))) return toupper(boardprint_getSymbolOfType(i));
    if (bitboard_test(b.black.pieces[i], SQUARE_FLATTEN(x, y))) return boardprint_getSymbolOfType(i);
  }
  return ' ';
}

static char transformSymbolBoardPrint(char symbol, int mode) {
  if (symbol != ' ') return symbol;
  else return mode ? '+' : ' ';
}

void boardprint_fen(const Board b) {
  char dest[100];
  fen_to_fen(b, dest);
  iolog_printf("%s", dest);
}

void boardprint_board(const Board b) {
  
  const char edge[] = "  -----------------\n";
  
  iolog_printf("%s", edge);
  int mode = 0;
  
  for (int y = 7; y >= 0; --y) {
    iolog_printf("%d| ", (y + 1));
    for (int x = 0; x < 8; ++x) {
      char s = getSymbol(b, x, y);
      char f = transformSymbolBoardPrint(s, mode);
      iolog_printf("%c ", f);
      
      
      mode = !mode;
    }
    mode = !mode;
    iolog_printf("|\n");
    
  }
  iolog_printf("%s", edge);
  iolog_printf("   A B C D E F G H\n");
  iolog_printf("\nfen: ");
  boardprint_fen(b);
  iolog_printf("\n");
  iolog_flush();
}

void boardprint_bitboard(const bitboard64 bits) {
  char boardString[64 + 64 + 8 + 1]; //bits + spaces + linebreaks + end
  char* ptr = &boardString[0];

  for (int j = 7; j >= 0; --j) {
    for (int i = 0; i < 8; ++i) {
      *ptr++ = bitboard_test(bits, i + j * 8) ? '1' : '0';
      *ptr++ = ' ';
    }
    *ptr++ = '\n';
  }
  *ptr = '\0';
  iolog_printf("%s\n", boardString);
}

