//
//  fen.c
//  fornax2_core
//
//  Created by Anders on 10/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//


#include <string.h>
#include <ctype.h>
#include "fen.h"
#include "../genericutils/require.h"
#include "../genericutils/strings.h"
#include "../boardprint.h"
#include "../board/board.h"
#include "../board/bitboard.h"
#include "../board/square.h"
#include "../tt/zobrist.h"
#include "../moving/movegen.h"


static char fen_getSymbol(Board b, Square sq) {
  int piece = board_get_inhabitant(&b, Square_flatten(sq), true);
  if (piece >= 0)  return toupper(boardprint_getSymbolOfType(piece));
  piece = board_get_inhabitant(&b, Square_flatten(sq), false);
  if (piece >= 0) return boardprint_getSymbolOfType(piece);
  return 0;
}

void fen_to_fen(const Board b, char* dest) {
  //  strcpy(dest, "abc\n");
  for (int j = 7; j >= 0; --j) {
    int spaceCounter = 0;
    for (int i = 0; i < 8; ++i) {
      
      Square sq = {i, j};
      char c = fen_getSymbol(b, sq);
      if (c > 0) {
        if (spaceCounter > 0) {
          (*dest++) = '0' + spaceCounter;
          spaceCounter = 0;
        }
        (*dest++) = c;
      }
      else {
        ++spaceCounter;
      }
    }
    if (spaceCounter > 0) {
      (*dest++) = '0' + spaceCounter;
    }
    (*dest++) = '/';
  }
  *(dest -1) = ' ';
  (*dest++) = b.isWhiteMove ? 'w' : 'b';
  (*dest++) = ' ';
  char* d = dest;
  bool whiteCastleK = b.isWhiteMove ? b.copyState->fck : b.copyState->hck;
  if (whiteCastleK) (*dest++) = 'K';
  bool whiteCastleQ = b.isWhiteMove ? b.copyState->fcq : b.copyState->hcq;
  if (whiteCastleQ) (*dest++) = 'Q';
  bool blackCastleK = b.isWhiteMove ? b.copyState->hck : b.copyState->fck;
  if (blackCastleK) (*dest++) = 'k';
  bool blackCastleQ = b.isWhiteMove ? b.copyState->hcq : b.copyState->fcq;
  if (blackCastleQ) (*dest++) = 'q';
  if (d == dest) (*dest++) = '-';
  
  (*dest++) = ' ';
  int pos = bitboard_findTrailing(b.copyState->enPassant);
  if (pos > 0) {
    (*dest++) = square_to_file(Square_unflatten(pos));
    (*dest++) = square_to_uci_rank(Square_unflatten(pos));
  }
  else (*dest++) = '-';
  
  (*dest++) = ' ';
  (*dest++) = '-';//todo
  (*dest++) = ' ';
  (*dest++) = '-';//todo
  *dest = 0;
}

void fen_to_board_split(Board* board, char** parts, int partsSize) {
  require(partsSize >= 4, "FEN was not comprised of at least 4 parts\n");
  
  char* fenRanks[8];
  int splitSizeBoard = strings_split(parts[0], fenRanks, "/");
  
  require(splitSizeBoard == 8, "FEN board was not comprised of 8 ranks: '%s'\n", parts[0]);
  
  board_clear(board);
  
  for (int j = 0; j < 8; ++j) {
    long size = strlen(fenRanks[j]);
    require(size <= 8 && size > 0, "FEN rank could not be parsed\n");
    int x = 0;
    int y = 7 - j;
    int counter = 0;
    while (counter < size) {
      char symbol = fenRanks[j][counter];
      
      switch (symbol) {
        case 'R':
          board_set_inhabitant(board, true, SQUARE_FLATTEN(x, y), INHABITANT_ROOKS);
          break;
        case 'N':
          board_set_inhabitant(board, true, SQUARE_FLATTEN(x, y), INHABITANT_KNIGHTS);
          break;
        case 'B':
          board_set_inhabitant(board, true, SQUARE_FLATTEN(x, y), INHABITANT_BISHOPS);
          break;
        case 'Q':
          board_set_inhabitant(board, true, SQUARE_FLATTEN(x, y), INHABITANT_QUEENS);
          break;
        case 'K':
          board_set_inhabitant(board, true, SQUARE_FLATTEN(x, y), INHABITANT_KINGS);
          break;
        case 'P':
          board_set_inhabitant(board, true, SQUARE_FLATTEN(x, y), INHABITANT_PAWNS);
          break;
        case 'r':
          board_set_inhabitant(board, false, SQUARE_FLATTEN(x, y), INHABITANT_ROOKS);
          break;
        case 'n':
          board_set_inhabitant(board, false, SQUARE_FLATTEN(x, y), INHABITANT_KNIGHTS);
          break;
        case 'b':
          board_set_inhabitant(board, false, SQUARE_FLATTEN(x, y), INHABITANT_BISHOPS);
          break;
        case 'q':
          board_set_inhabitant(board, false, SQUARE_FLATTEN(x, y), INHABITANT_QUEENS);
          break;
        case 'k':
          board_set_inhabitant(board, false, SQUARE_FLATTEN(x, y), INHABITANT_KINGS);
          break;
        case 'p':
          board_set_inhabitant(board, false, SQUARE_FLATTEN(x, y), INHABITANT_PAWNS);
          break;
          
        default:
        {
          int range = symbol - '0';
          
          if (range > 0 && range <= 8) x += range - 1;
          else require(false, "FEN unknown symbol: %c\n", symbol);
        }
      };
      counter++;
      x++;
    }
  }
  {
    if (parts[1][0] == 'w') board->isWhiteMove = true;
    else if (parts[1][0] == 'b') board->isWhiteMove = false;
    else require(false, "FEN unknown symbol: %s\n", parts[1]);
  }
  {
    long size = strlen(parts[2]);
    for (int i = 0; i < size; ++i) {
      if (parts[2][i] == '-') break;
      else if (parts[2][i] == 'K') {
        if (board->isWhiteMove) board->copyState->fck = 1;
        else board->copyState->hck = 1;
      }
      else if (parts[2][i] == 'Q') {
        if (board->isWhiteMove) board->copyState->fcq = 1;
        else board->copyState->hcq = 1;
      }
      else if (parts[2][i] == 'k') {
        if (board->isWhiteMove) board->copyState->hck = 1;
        else board->copyState->fck = 1;
      }
      else if (parts[2][i] == 'q') {
        if (board->isWhiteMove) board->copyState->hcq = 1;
        else board->copyState->fcq = 1;
      }
      else require(false, "FEN unknown symbol(s): %s\n", parts[2]);
    }
    if (parts[3][0] != '-') bitboard_set(&board->copyState->enPassant, Square_flatten(square_from_uci(parts[3])));
    if (partsSize > 4 && parts[4][0] != '-')
      board->copyState->halfMoveCount = (uint8_t) string_to_long(parts[4]);
//    if (partsSize > 5 && parts[5][0] != '-')
      //todo
  }
  
  board_set_friendlies_and_hostiles(board);
  board->copyState->hash = zobrist_get(board);
  movegen_genmaps(board);
  movegen_genmaps_hostile_attacks(board);
  
  assert( (board->white.pieces[0]
         & board->white.pieces[1]
         & board->white.pieces[2]
         & board->white.pieces[3]
         & board->white.pieces[4]
         & board->white.pieces[5]) == 0);
  assert( (board->black.pieces[0]
         & board->black.pieces[1]
         & board->black.pieces[2]
         & board->black.pieces[3]
         & board->black.pieces[4]
         & board->black.pieces[5]) == 0);
}

void fen_to_board(Board* board, const char* fen) {
  char fenCopy[100];
  strcpy(fenCopy, fen);
  
  char* parts[6];
  
  int splitSize = strings_split(fenCopy, parts, " ");
  require(splitSize >= 4, "FEN was not comprised of at least 4 parts: '%s'\n", fen);
  fen_to_board_split(board, parts, splitSize);
}
