//
//  movegenerator.c
//  fornax2_core
//
//  Created by Anders on 30.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#include "movegen.h"
#include "oracle.h"
#include "../board/bitboard.h"
#include "../board/board.h"
#include "../board/definitions.h"
#include "../board/square.h"
#include "../board/move.h"
#include "movegen_pawns.inc.h"

#ifndef NDEBUG
#include <stdio.h>
#include "../boardprint.h"
#endif

static inline void moveset_register_attacks(BoardSide* side,
                                            inhabitant piece,
                                            bitboard64 attacks) {
//  if (attacks == 0) return;
  side->attackMask |= attacks;
  Attackmap set = {piece, attacks};
  assert(side->attackmapCount < 16);//up to 32 is supported
  side->attackmaps[side->attackmapCount++] = set;
}

static inline void moveset_register_moves(Board* board, square origin, inhabitant piece, bitboard64 moves) {
//  if (moves == 0) return;
  Movemap set = {origin, piece, moves};
  assert(board->movemapCount < 17);//up to 32 is supported
  board->movemaps[board->movemapCount++] = set;
}

static inline void moveset_castle(Board* board) {
  bitboard64 moves = 0;
  bitboard64 total = board->white.pieceMask | board->black.pieceMask;
  {
    bitboard64 castlingBlockers = oracle_get_castling_blocker(true, board->isWhiteMove);
    bool canCastleK = board->copyState->fck & !(castlingBlockers & total);
    moves |= ((bitboard64)0 - canCastleK) & oracle_get_castling_dest(true, board->isWhiteMove);
  }
  {
    bitboard64 castlingBlockers = oracle_get_castling_blocker(false, board->isWhiteMove);
    bool canCastleQ = board->copyState->fcq  & !(castlingBlockers & total);
    moves |= ((bitboard64)0 - canCastleQ) & oracle_get_castling_dest(false, board->isWhiteMove);
  }
  square origin = moves > (68) ? 60 : 4;
  moveset_register_moves(board, origin, INHABITANT_KINGS, moves);
}


static inline void moveset_pawn_attacks(BoardSide* side, bool isWhite) {
  bitboard64 pawns = side->pieces[INHABITANT_PAWNS];
  bitboard64 attacks = isWhite
  ? BITBOARD_SHIFT_N(pawns)
  : BITBOARD_SHIFT_S(pawns);
  bitboard64 east = BITBOARD_SHIFT_E(attacks);
  moveset_register_attacks(side, INHABITANT_PAWNS, east);
  bitboard64 west = BITBOARD_SHIFT_W(attacks);
  moveset_register_attacks(side, INHABITANT_PAWNS, west);
}

static inline bitboard64 rook_slide(square pos, bitboard64 all) {
  BITBOARD_CLEAR(all, pos);
  bitboard64 attacks;
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_N);
    bitboard64 b = BITBOARD_SHIFT_N(all) & ray;
    b = bitboard_fill_N(b);
    attacks = (~b & ray);
  }
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_E);
    bitboard64 b = BITBOARD_SHIFT_E(all) & ray;
    b = bitboard_fill_E(b);
    attacks |= (~b & ray);
  }
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_S);
    bitboard64 b = BITBOARD_SHIFT_S(all) & ray;
    b = bitboard_fill_S(b);
    attacks |= (~b & ray);
  }
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_W);
    bitboard64 b = BITBOARD_SHIFT_W(all) & ray;
    b = bitboard_fill_W(b);
    attacks |= (~b & ray);
  }
  return attacks;
}

static inline bitboard64 bishop_slide(square pos, bitboard64 all) {
  BITBOARD_CLEAR(all, pos);
  bitboard64 attacks;
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_NW);
    bitboard64 b = BITBOARD_SHIFT_NW(all) & ray;
    b = bitboard_fill_NW(b);
    attacks = (~b & ray);
  }
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_NE);
    bitboard64 b = BITBOARD_SHIFT_NE(all) & ray;
    b = bitboard_fill_NE(b);
    attacks |= (~b & ray);
  }
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_SE);
    bitboard64 b = BITBOARD_SHIFT_SE(all) & ray;
    b = bitboard_fill_SE(b);
    attacks |= (~b & ray);
  }
  {
    bitboard64 ray = oracle_get_sliding_attacks(pos, DIR_SW);
    bitboard64 b = BITBOARD_SHIFT_SW(all) & ray;
    b = bitboard_fill_SW(b);
    attacks |= (~b & ray);
  }
  return attacks;
}

static inline void moveset_bishops(Board* board, bitboard64 bishops, inhabitant piecetype) {
  bitboard64 friendlies = board->friendly->pieceMask;
  bitboard64 hostiles = board->hostile->pieceMask;
  bitboard64 all = friendlies | hostiles;
  
  square pos;
  while ((pos = bitboard_popTrailing(&bishops)) != SQUARE_NONE) {
    assert(SQUARE_INBOUNDS(pos));
    
    bitboard64 attacks = bishop_slide(pos, all);
    
    moveset_register_attacks(board->friendly, piecetype, attacks);
    attacks &= ~friendlies;
    moveset_register_moves(board, pos, piecetype, attacks);
  }
}

static inline void moveset_rooks(Board* board, bitboard64 rooks, inhabitant piecetype) {
  bitboard64 friendlies = board->friendly->pieceMask;
  bitboard64 hostiles = board->hostile->pieceMask;
  bitboard64 all = friendlies | hostiles;
  
  square pos;
  while ((pos = bitboard_popTrailing(&rooks)) != SQUARE_NONE) {
    assert(SQUARE_INBOUNDS(pos));
    
    bitboard64 attaks = rook_slide(pos, all);
    moveset_register_attacks(board->friendly, piecetype, attaks);
    attaks &= ~friendlies;
    moveset_register_moves(board, pos, piecetype, attaks);
  }
}

static inline void moveset_knights(Board* board, bitboard64 knights) {
  
  square pos;
  while ((pos = bitboard_popTrailing(&knights)) != SQUARE_NONE) {
    assert(SQUARE_INBOUNDS(pos));
    bitboard64 attacks = oracle_get_knight_attacks(pos);
    moveset_register_attacks(board->friendly, INHABITANT_KNIGHTS, attacks);
    attacks &= ~board->friendly->pieceMask;
    moveset_register_moves(board, pos, INHABITANT_KNIGHTS, attacks);
  }
}

static inline void moveset_kings(Board* board, bitboard64 king) {
  const square kingPosition = BITBOARD_FIND_TRAILING(king);
  assert(SQUARE_INBOUNDS(kingPosition));
  bitboard64 attacks = oracle_get_king_attacks(kingPosition);
  moveset_register_attacks(board->friendly, INHABITANT_KINGS, attacks);
  attacks &= ~board->friendly->pieceMask;
  moveset_register_moves(board, kingPosition, INHABITANT_KINGS, attacks);
}

void movegen_genmaps(Board* board) {
  board->movemapCount = 0;
  board->friendly->attackmapCount = 0;
  board->friendly->attackMask = 0;
  bitboard64* array = board->friendly->pieces;
  
  moveset_kings(board, *array);
  moveset_bishops(board, *(array + INHABITANT_QUEENS), INHABITANT_QUEENS);
  moveset_rooks(board, *(array + INHABITANT_QUEENS), INHABITANT_QUEENS);
  moveset_rooks(board, *(array + INHABITANT_ROOKS), INHABITANT_ROOKS);
  moveset_bishops(board, *(array + INHABITANT_BISHOPS), INHABITANT_BISHOPS);
  moveset_knights(board, *(array + INHABITANT_KNIGHTS));
  moveset_castle(board);
  
  moveset_pawn_attacks(board->friendly, board->isWhiteMove);
}

move* movegen_set_to_moves(Board* board, move* pointer, Movemap* set) {
  inhabitant piece = set->piece;
  square origin = set->origin;
  
  move m = 0;
  m = move_set_origin_piece(m, piece);
  m = move_set_origin(m, origin);
  
  const piecetable* table = board->isWhiteMove ? &WHITE_MOVE_PIECETABLE : &BLACK_MOVE_PIECETABLE;
  
  int8_t basevalue = -((*table)[piece][origin]);
    m = move_set_destination_piece(m, piece);
    square dest;
    while ((dest = bitboard_popTrailing(&set->moves)) != SQUARE_NONE) {
      m = move_set_destination(m, dest);
      inhabitant capture = board->capturemap[dest];
      assert(capture == board_get_captured(board, dest));
      m = move_set_capture(m, capture);
      int8_t captureValue = basevalue + MOVE_CAPTUREVALUE[capture];
      bool isProtectedSquare = BITBOARD_TEST(board->hostile->attackMask, dest);
      int8_t posvalue = captureValue + move_destination_penalty_value(piece, isProtectedSquare) + ((*table)[piece][dest]);
      
      move final = move_set_value(m, posvalue);
      *pointer++ = final;
    }
  return pointer;
}

uint8_t movegen_gen(Board* board, move* dest_size256) {
  move* pointer = dest_size256;
  
  for (int i = 0; i < board->movemapCount; ++i) {
    Movemap* set = &board->movemaps[i];
    pointer = movegen_set_to_moves(board, pointer, set);
  }
  
  pointer = movegen_gen_pawn_moves(board, pointer);
  
  uint8_t count = pointer - dest_size256;
    
  #ifndef NDEBUG
    for (int i = 0; i < count; ++i) {
      move m = dest_size256[i];
      int8_t valWas = MOVE_VALUE(m);
      int8_t valShould = MOVE_VALUE(move_annotate_order_value(m, board));
      if (valWas != valShould) {
        boardprint_debugmove(m);
        printf("%d vs %d\n", valWas, valShould);
      }
      assert(valWas == valShould);
    }
  #endif
    
  return count;
}

void movegen_genmaps_hostile_attacks(Board* board) {
  BoardSide* hostile = board->hostile;
  hostile->attackmapCount = 0;
  hostile->attackMask = 0;
  
  bitboard64 all = board->white.pieceMask | board->black.pieceMask;
  bitboard64* hostiles = hostile->pieces;
  
  square pos;
  moveset_pawn_attacks(board->hostile, !board->isWhiteMove);
  
  bitboard64 knighs = hostiles[INHABITANT_KNIGHTS];
  while ((pos = bitboard_popTrailing(&knighs)) != SQUARE_NONE) {
    bitboard64 attacks = oracle_get_knight_attacks(pos);
    moveset_register_attacks(hostile, INHABITANT_KNIGHTS, attacks);
  }
  bitboard64 bishops = hostiles[INHABITANT_BISHOPS];
  while ((pos = bitboard_popTrailing(&bishops)) != SQUARE_NONE) {
    bitboard64 attacks = bishop_slide(pos, all);
    moveset_register_attacks(hostile, INHABITANT_BISHOPS, attacks);
  }
  bitboard64 rooks = hostiles[INHABITANT_ROOKS];
  while ((pos = bitboard_popTrailing(&rooks)) != SQUARE_NONE) {
    bitboard64 attacks = rook_slide(pos, all);
    moveset_register_attacks(hostile, INHABITANT_ROOKS, attacks);
  }
  bitboard64 queens = hostiles[INHABITANT_QUEENS];
  while ((pos = bitboard_popTrailing(&queens)) != SQUARE_NONE) {
    bitboard64 attacks = bishop_slide(pos, all);
    attacks |= rook_slide(pos, all);
    moveset_register_attacks(hostile, INHABITANT_QUEENS, attacks);
  }
  pos = BITBOARD_FIND_TRAILING(hostiles[INHABITANT_KINGS]);
  bitboard64 kingAttacks = oracle_get_king_attacks(pos);
  moveset_register_attacks(hostile, INHABITANT_KINGS, kingAttacks);
  
  board->isCheck = board->friendly->pieces[INHABITANT_KINGS] & board->hostile->attackMask;
}
