//
//  oracle.c
//  fornax2_core
//
//  Created by Anders on 30.06.2018.
//  Copyright © 2018 Avividi. All rights reserved.
//

#include <stdlib.h>

#include "oracle.h"
#include "../board/bitboard.h"
#include "../board/square.h"
#include "../boardprint.h"

bitboard64* oracle_king_attacks;
bitboard64* oracle_king_neighbor_vuln_squares;
bitboard64* oracle_knight_attacks;
bitboard64 (*oracle_sliding_attaks)[64];
bitboard64* oracle_castling_dest_and_blocks;
bitboard64* oracle_position_to_bitboard;
_Bool hasGenerated = 0;

static bitboard64 oracle_preCalcuateKingNeighborSquares(int position) {
  bitboard64 neighbors = oracle_get_king_attacks(position);
  bitboard_set(&neighbors, position);
  
  if (SQUARE_Y(position) > 3) neighbors |= neighbors >> 8;
  else neighbors |= neighbors << 8;
  
  return neighbors;
}

static bitboard64 oracle_preCalculateKingAttaks(int position) {
  bitboard64 attaks = 0;
  Square origin = Square_unflatten(position);
  
  for (int i = 0; i < 8; i++) {
    Square dest = Square_add(origin, CARDINALS[i]);
    if (Square_inBounds(dest)) bitboard_set(&attaks, SQUARE_FLATTEN(dest.x, dest.y));
  }
  
  return attaks;
}

static bitboard64 oracle_preCalculateKnightAttaks(int position) {
  bitboard64 attaks = 0;
  
  Square origin =  Square_unflatten(position);
  const Square knightMoves[8] = {
    {1, -2}, {-1, -2}, {-2, -1}, {-2, 1}, {-1, 2}, {1, 2}, {2, 1}, {2, -1}
  };
  
  for (int i = 0; i < 8; i++) {
    Square dest = Square_add(origin, knightMoves[i]);
    if (Square_inBounds(dest)) bitboard_set(&attaks, SQUARE_FLATTEN(dest.x, dest.y));
  }
  
  return attaks;
}

static void oracle_preCalculateSlidingAttaks(int position) {
  Square start = Square_unflatten(position);
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (++sq.y < 8) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_N][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (++sq.y < 8 && ++sq.x < 8) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_NE][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (++sq.x < 8) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_E][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (--sq.y >= 0 && ++sq.x < 8) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_SE][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (--sq.y >= 0) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_S][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (--sq.y >= 0 && --sq.x >= 0) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_SW][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (--sq.x >= 0) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_W][position] = ray;
  }
  {
    Square sq = {start.x, start.y};
    bitboard64 ray = 0;
    while (++sq.y < 8 && --sq.x >= 0) bitboard_set(&ray, SQUARE_FLATTEN(sq.x, sq.y));
    oracle_sliding_attaks[DIR_NW][position] = ray;
  }
}

void oracle_init(void) {
  oracle_king_attacks = malloc(sizeof(*oracle_king_attacks) * 64);
  oracle_king_neighbor_vuln_squares = malloc(sizeof(*oracle_king_neighbor_vuln_squares) * 64);
  oracle_knight_attacks = malloc(sizeof(*oracle_knight_attacks) * 64);
  oracle_sliding_attaks = malloc(sizeof(*oracle_sliding_attaks) * 8);
  for (int i = 0; i < 64; i++) {
    oracle_king_attacks[i] = oracle_preCalculateKingAttaks(i);
    oracle_knight_attacks[i] = oracle_preCalculateKnightAttaks(i);
    oracle_preCalculateSlidingAttaks(i);
    oracle_king_neighbor_vuln_squares[i] = oracle_preCalcuateKingNeighborSquares(i);
  }
  oracle_castling_dest_and_blocks = malloc(sizeof(*oracle_castling_dest_and_blocks) * 8);
  for (int i = 0; i < 8; ++i) {
    oracle_castling_dest_and_blocks[i] = 0;
  }
  bitboard_set(&oracle_castling_dest_and_blocks[0], SQUARE_FLATTEN(2, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[1], SQUARE_FLATTEN(2, 0));
  bitboard_set(&oracle_castling_dest_and_blocks[2], SQUARE_FLATTEN(6, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[3], SQUARE_FLATTEN(6, 0));
  bitboard_set(&oracle_castling_dest_and_blocks[4], SQUARE_FLATTEN(1, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[4], SQUARE_FLATTEN(2, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[4], SQUARE_FLATTEN(3, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[5], SQUARE_FLATTEN(1, 0));
  bitboard_set(&oracle_castling_dest_and_blocks[5], SQUARE_FLATTEN(2, 0));
  bitboard_set(&oracle_castling_dest_and_blocks[5], SQUARE_FLATTEN(3, 0));
  bitboard_set(&oracle_castling_dest_and_blocks[6], SQUARE_FLATTEN(5, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[6], SQUARE_FLATTEN(6, 7));
  bitboard_set(&oracle_castling_dest_and_blocks[7], SQUARE_FLATTEN(5, 0));
  bitboard_set(&oracle_castling_dest_and_blocks[7], SQUARE_FLATTEN(6, 0));
  
  
  oracle_position_to_bitboard = malloc(sizeof(*oracle_position_to_bitboard) * 64);
  for (int i = 0; i < 64; ++i) {
    oracle_position_to_bitboard[i] = 0;
    bitboard_set(&oracle_position_to_bitboard[i], i);
  }
  
  hasGenerated = true;

}

void oracle_destroy(void) {
  free(oracle_king_attacks);
  free(oracle_king_neighbor_vuln_squares);
  free(oracle_knight_attacks);
  free(oracle_sliding_attaks);
  free(oracle_castling_dest_and_blocks);
  free(oracle_position_to_bitboard);
}

