//
//  search.c
//  fornax2_core
//
//  Created by Anders on 28/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#include <stdio.h>
#include <inttypes.h>
#include "search.h"
#include "../tt/transpositions.h"
#include "../moving/movegen.h"
#include "../moving/makemove.h"
#include "../moving/movepicker.h"
#include "../boardprint.h"
#include "../board/board.h"
#include "../board/move.h"
#include "../eval/eval.h"
#include "../genericutils/stopwatch.h"
#include "../genericutils/debugprint.h"
#include "../genericutils/iolog.h"
#include "../supplementary/time_management.h"

#define CHECK_STOP if (*stopflag) { return 0; }
#define CURRMOVE_NODE_THRESHOLD 2500000

static bool* stopflag;
static uint64_t nodeCount = 0;
static uint64_t betacuts = 0;
static uint64_t nullcuts = 0;
static uint64_t quiesceCount = 0;
static uint64_t transpositionHits = 0;

static long getnps(long duration) {
  long time = duration | 1;
  return ((nodeCount | 1) / time) * 1000;
}

static void printline(const Board* board, uint8_t depth, eval eval, move move, long duration) {
  iolog_printf("info depth %d score cp %d nodes %" PRIu64 " nps %ld", depth, eval, nodeCount, getnps(duration));
#if DEBUG
  iolog_printf(" tthits %" PRIu64, transpositionHits);
  iolog_printf(" qnodes %" PRIu64, quiesceCount);
  iolog_printf(" betacuts %" PRIu64, betacuts);
  iolog_printf(" nullcuts %" PRIu64, nullcuts);
#endif
  iolog_printf(" pv ");
  boardprint_printpv(board, move, depth);
  iolog_printf("\n");
  iolog_flush();
}

static void printcurrmove(uint8_t depth, move m, uint8_t count) {
  iolog_printf("info depth %d currmove ", depth);
  boardprint_move(m);
  iolog_printf(" currmovenumber %d\n", count);
  iolog_flush();
}

void search_printinfo(void) {
  printf("info nodes %" PRIu64 " quiesce %" PRIu64 " tthits %" PRIu64 "\n", nodeCount, quiesceCount, transpositionHits);
}

static inline eval probe_tt(ttkey hash, move* ttmove, int depth, eval alpha, eval beta) {
  Bucket* b = transpositions_get(hash);
  if (transpositions_age(b) != 0) {
    *ttmove = b->move;
    
    transpositionHits++;

    if (b->depth >= depth) {
      tttype type = transpositions_type(b);
      if(type == TT_TYPE_EXACT) return b->eval;
      if(type == TT_TYPE_ALPHA && b->eval <= alpha) return alpha;
      if(type == TT_TYPE_BETA && b->eval >= beta) return beta;
    }
  }
  
  return EVAL_NONE;
}

static inline eval quiesce_check(Board* node, uint8_t depth, eval alpha, eval beta);

static eval quiesce(Board* node, uint8_t depth, eval alpha, eval beta) {
  
  quiesceCount++;
  nodeCount++;
  
  if (node->isCheck) {
    return quiesce_check(node, depth, alpha, beta);
  }
  
  eval pat = eval_evaluate(node);
  if (depth == 0) return pat;
  if (pat >= beta) return beta;
  if (pat > alpha) alpha = pat;

  BoardCopiedState* preservedState = node->copyState;
  
  Movepick state;
  movepick_start_quiesce(&state, node);
  move m;
  while ((m = movepick_next_quiesce(&state)) != 0) {
    CHECK_STOP;
    
    BoardCopiedState nextstate;
    bool legal = makemove_make(node, &nextstate, m);
    if (!legal) {
      makemove_takeback(node, preservedState, m);
      continue;
    }
    eval eval = -quiesce(node, depth - 1, -beta, -alpha);
    makemove_takeback(node, preservedState, m);
    if (eval >= beta) {
      betacuts++;
      return beta;
    }
    if (eval > alpha) {
      alpha = eval;
    }
  }
  return alpha;
}

static inline eval quiesce_check(Board* node, uint8_t depth, eval alpha, eval beta) {
  move ttmove = 0;
  int tteval;
  if ((tteval = probe_tt(node->copyState->hash, &ttmove, depth, alpha, beta)) != EVAL_NONE) {
    return tteval;
  }
  
  bool nolegal = true;
  const bool isLeaf = depth == 0;
  
  eval pat = EVAL_NONE;
  if (isLeaf) {
    pat = eval_evaluate(node);
  }
  
  BoardCopiedState* preservedState = node->copyState;
  
  Movepick state;
  movepick_start(&state, node, ttmove);
  
  move m;
  while ((m = movepick_next(&state)) != 0) {
    CHECK_STOP;
    
    BoardCopiedState nextstate;
    bool legal = makemove_make(node, &nextstate, m);
    
    if (!legal) {
      makemove_takeback(node, preservedState, m);
      continue;
    }
    else if (isLeaf) {
      makemove_takeback(node, preservedState, m);
      return pat;
    }
    
    nolegal = false;
    
    eval eval = -quiesce(node, depth - 1, -beta, -alpha);
    makemove_takeback(node, preservedState, m);
    if (eval >= beta) {
      betacuts++;
      return beta;
    }
    if (eval > alpha) {
      alpha = eval;
    }
  }
  
  if (nolegal) {
    return (EVAL_CHECKMATE + 1);
  }
  
  return alpha;
}

static eval search_node(Board* board, bool checkNull, uint8_t depth, eval alpha, eval beta);

static uint8_t late_move_reduction(uint8_t depth, move m, bool isCheck, bool isCheckingMove) {
  if (!isCheck
      && !isCheckingMove
      && depth > 2
      && MOVE_ORIGIN_PIECE(m) != INHABITANT_PAWNS
      && !MOVE_CAPTURE(m)) {
//    int8_t movevalue = MOVE_VALUE(m);
//    if (movevalue < MOVE_BASEVALUE - 10) return 2 + depth / 4;
//    else
     return 1;
  }
  return 0;
}

static inline eval null_search(Board* node, uint8_t depth, eval beta) {
  BoardCopiedState* preservedState = node->copyState;
  BoardCopiedState newState;
  makemove_nullmove_make(node, &newState);
  uint8_t reduced = depth - 2;
  reduced -= depth / 5;
  eval eval = -search_node(node, false, reduced - 1, -beta, -beta + 1);
  makemove_nullmove_takeback(node, preservedState);
  return eval;
}

static eval search_node(Board* node, bool checkNull, uint8_t depth, eval alpha, eval beta) {
  if (depth == 0) {
    return quiesce(node, 5, alpha, beta);
  }
  assert(depth > 0);
  nodeCount++;
  
  if (board_is_repetition(node)) return EVAL_DRAW;
  
  move ttmove = 0;
  eval tteval;
  if ((tteval = probe_tt(node->copyState->hash, &ttmove, depth, alpha, beta)) != EVAL_NONE) {
    return tteval;
  }
  
  bool isCheck = node->isCheck;
  
  if (isCheck) {
    if (node->hostile->eval_material > MATERIAL_VALUE_EXTEND_CHECK_THRESHOLD) {
      depth++;
    }
  }
  else if (checkNull && depth > 2 && node->friendly->eval_material > MATERIAL_VALUE_NULL_THRESHOLD) {
    eval eval = null_search(node, depth, beta);
    if(eval >= beta) {
      betacuts++;
      nullcuts++;
      return beta;
    }
    movegen_genmaps(node);
  }
  
  eval criticalEval = EVAL_NONE;
  move criticalMove = 0;
  tttype tttype = TT_TYPE_ALPHA;
  
  BoardCopiedState* preservedState = node->copyState;
  
  Movepick state;
  movepick_start(&state, node, ttmove);
  move m;
  while ((m = movepick_next(&state)) != 0) {
    CHECK_STOP;
    
    BoardCopiedState nextstate;
    bool legal = makemove_make(node, &nextstate, m);
    if (!legal) {
      makemove_takeback(node, preservedState, m);
      continue;
    }
    
    if (criticalEval == EVAL_NONE) {
      criticalEval = -search_node(node, true, depth - 1, -beta, -alpha);
      makemove_takeback(node, preservedState, m);
      criticalMove = m;
      
      if (criticalEval >= beta) {
        betacuts++;
        transpositions_put(node->copyState->hash, depth, criticalEval, criticalMove, TT_TYPE_BETA);
        movepick_set_killer(node->ply, m, ttmove);
        return criticalEval;
      }
      if (criticalEval > alpha) {
        alpha = criticalEval;
        tttype = TT_TYPE_EXACT;
      }
      continue;
    }
    
    uint8_t reduction = late_move_reduction(depth, m, isCheck, node->isCheck);
    
    eval eval = -search_node(node, true, depth - reduction - 1, -alpha - 1, -alpha);
    makemove_takeback(node, preservedState, m);
    
    if (eval > alpha && eval < beta) {
      makemove_make(node, &nextstate, m);
      eval = -search_node(node, true, depth - 1, -beta, -alpha);
      makemove_takeback(node, preservedState, m);
    }
    if (eval > criticalEval) {
      criticalEval = eval;
      criticalMove = m;
    }
    if (eval >= beta) {
      betacuts++;
      transpositions_put(node->copyState->hash, depth, eval, criticalMove, TT_TYPE_BETA);
      movepick_set_killer(node->ply, m, ttmove);
      return eval;
    }
    if (eval > alpha) {
      alpha = eval;
      tttype = TT_TYPE_EXACT;
    }
  }
  
  if (criticalEval == EVAL_NONE) {
    criticalEval = isCheck ? (EVAL_CHECKMATE - depth) : EVAL_DRAW;
    tttype = TT_TYPE_EXACT;
  }
  transpositions_put(node->copyState->hash, depth, criticalEval, criticalMove, tttype);
  return criticalEval;
}

static void search_set_results(SearchResult* result, const Board* root, uint8_t depth, eval eval, move move, long starttime) {
  long duration = stopwatch_wall_stop_instance(starttime);
  printline(root, depth, eval, move, duration);
  result->move = move;
  result->eval = eval;
  result->duration_ms= duration;
}

static eval search_root(SearchResult* result,
                        RootMovePick* rootmoves,
                        const Board* root,
                        uint8_t depth,
                        eval alpha,
                        eval beta,
                        bool printcurr) {
  
  assert(depth > 0);
  nodeCount = 1;
  transpositionHits = 0;
  quiesceCount = 0;
  betacuts = 0;
  nullcuts = 0;
  
  long starttime = stopwatch_wall_start_instance();
  
  move ttmove = 0;
  eval tteval;
  if ((tteval = probe_tt(root->copyState->hash, &ttmove, depth, alpha, beta)) != EVAL_NONE) {
    search_set_results(result, root, depth, tteval, ttmove, starttime);
    if (!board_is_repetition(root)) return tteval;
  }
  
  Board board = BOARD_COPY(board, root);
  Board* node = &board;
  
  bool isCheck = node->isCheck;
  
  eval criticalEval = EVAL_NONE;
  move criticalMove = 0;
  tttype tttype = TT_TYPE_ALPHA;
  
  BoardCopiedState* preservedState = node->copyState;
  
//    Movepick state;
//    movepick_start(&state, node, ttmove);
//    move m;
//    while ((m = movepick_next(&state)) != 0) {
  movepick_root_list_reset(rootmoves, ttmove);
  move m;
  uint8_t currmovenumber = 0;
  while ((m = movepick_root_next(rootmoves)) != 0) {
    CHECK_STOP;
    
    currmovenumber++;
    if (printcurr) {
      printcurrmove(depth, m, currmovenumber);
    }
    
    BoardCopiedState nextstate;
    bool legal = makemove_make(node, &nextstate, m);
    if (!legal) {
      makemove_takeback(node, preservedState, m);
      continue;
    }
    
    if (criticalEval == EVAL_NONE) {
      criticalEval = -search_node(node, true, depth - 1, -beta, -alpha);
      makemove_takeback(node, preservedState, m);
      criticalMove = m;
      
      if (criticalEval >= beta) {
        betacuts++;
        transpositions_put(node->copyState->hash, depth, criticalEval, criticalMove, TT_TYPE_BETA);
        return criticalEval;
      }
      if (criticalEval > alpha) {
        alpha = criticalEval;
        tttype = TT_TYPE_EXACT;
      }
      continue;
    }

    uint8_t reduction = late_move_reduction(depth, m, isCheck, node->isCheck);
//    uint8_t reduction = 0;
    
    eval eval = -search_node(node, true, depth - reduction - 1, -alpha - 1, -alpha);
    makemove_takeback(node, preservedState, m);
    
    if (eval > alpha && eval < beta) {
      makemove_make(node, &nextstate, m);
      eval = -search_node(node, true, depth - 1, -beta, -alpha);
      makemove_takeback(node, preservedState, m);
    }
    if (eval > criticalEval) {
      criticalEval = eval;
      criticalMove = m;
    }
    if (eval >= beta) {
      betacuts++;
      transpositions_put(node->copyState->hash, depth, eval, criticalMove, TT_TYPE_BETA);
      return eval;
    }
    if (eval > alpha) {
      alpha = eval;
      tttype = TT_TYPE_EXACT;
    }
  }
  if (criticalEval == EVAL_NONE) {
    criticalEval = isCheck ? (EVAL_CHECKMATE - depth) : EVAL_DRAW;
    tttype = TT_TYPE_EXACT;
  }
  transpositions_put(node->copyState->hash, depth, criticalEval, criticalMove, tttype);
  
  if (!*stopflag) {
    search_set_results(result, node, depth, criticalEval, criticalMove, starttime);
  }
  
//  transpositions_maintenance();
  return criticalEval;
}


static void root_move_populate(RootMovePick* rootmoves, const Board* board) {
  Board root = BOARD_COPY(root, board);
  move moves[256];
  uint8_t moveCount = movegen_gen(&root, moves);
  
  BoardCopiedState* prevstate = root.copyState;
  
  for (uint8_t i = 0; i < moveCount; ++i) {
    move m = moves[i];
    
    BoardCopiedState nextstate;
    bool legal = makemove_make(&root, &nextstate, m);
    if (legal) {
      eval q = - quiesce(&root, 5, EVAL_ALPHASTART, EVAL_BETASTART);
      movepick_root_list_add(rootmoves, m, q);
    }
    makemove_takeback(&root, prevstate, m);
  }
}

move search_iterative_stop(const Board* board, uint8_t depth, bool* stop) {
  stopflag = stop;
  uint8_t currdepth = 1;
  bool printcurr = false;
  
  eval alpha = EVAL_ALPHASTART;
  eval beta = EVAL_BETASTART;
  
  SearchResult result = {.move = 0, .eval = EVAL_NONE, .duration_ms = 0};
  
  movepick_clean_killers();
  
  RootMovePick* rootmoves = movepick_root_list_init();
  root_move_populate(rootmoves, board);
  
  while (currdepth <= depth) {
    search_root(&result, rootmoves, board, currdepth, alpha, beta, printcurr);
    currdepth++;
    printcurr = nodeCount > CURRMOVE_NODE_THRESHOLD;
    
    if (currdepth == 0) break;
    if (*stopflag) break;
    if (time_management_check_stop()) break;
  }
  time_management_send_stopped();
  boardprint_printbest(result.move);
  
  movepick_root_list_destroy(rootmoves);
  return result.move;
}


move search_iterative(const Board* board, uint8_t depth) {
  bool stopflag = false;
  return search_iterative_stop(board, depth, &stopflag);
}
