//
//  opening_book.c
//  fornax2_core
//
//  Created by Anders on 28/03/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#include <stdlib.h>
#include <time.h>
#include <stdarg.h>
#include "opening_book.h"
#include "../genericutils/require.h"
#include "../fen/fen.h"
#include "../board/move.h"
#include "../moving/movegen.h"
#include "../moving/makemove.h"

#define BOOK_SIZE 20
#define MOVES_SIZE 40
static bool disabled = true;

typedef struct {
  uint64_t ttkey;
  move m[MOVES_SIZE];
  uint8_t count;
} Opening;

static Opening* openings;

Opening* opening_book_find(const Board* board) {
  int i = 0;
  Opening* opening;
  while ((opening = &openings[i++])->ttkey != 0) {
    
    if (opening->ttkey == board->copyState->hash) {
      return opening;
    }
  }
  return NULL;
}

static bool isLegalMove(const Board* board, move m) {
  Board copy = BOARD_COPY(copy, board);
  move moves[256];
  uint8_t moveCount = movegen_gen(&copy, moves);
  for (uint8_t i = 0; i < moveCount; ++i) {
    if (moves[i] != m) continue;
    BoardCopiedState nextstate;
    return makemove_make(&copy, &nextstate, m);
  }
  return false;
}

static void opening_register(const Board* board, const char* ucimove) {
  
  Board b = BOARD_COPY(b, board);
  move m = move_full_from_uci(&b, ucimove, 4);
  
  require(isLegalMove(&b, m), "illegal move %s\n", ucimove);
  
  Opening* opening = opening_book_find(&b);
  if (opening != NULL) {
    require(opening->count < MOVES_SIZE, "too many moves for this position, size=%d\n", MOVES_SIZE);
    opening->m[opening->count] = m;
    opening->count++;

    return;
  }
  
  int i = 0;
  while ((openings[i]).ttkey != 0) i++;
  require(i < BOOK_SIZE, "book was full, size=%d\n", BOOK_SIZE);
  Opening o;
  o.ttkey = b.copyState->hash;
  o.m[0] = m;
  o.count = 1;
  openings[i] = o;
}

static void opening_moves_to_board(Board* board, ...) {
  
  const char* ucimove;
  BoardCopiedState* returnstate = board->copyState;
  
  va_list arglist;
  va_start(arglist, board);
  
  while ((ucimove = va_arg(arglist, const char *)) != NULL) {
    move m = move_full_from_uci(board, ucimove, 4);
    require(isLegalMove(board, m), "illegal move %s\n", ucimove);
    BoardCopiedState nextstate;
    makemove_make(board, &nextstate, m);

    *returnstate = *board->copyState;
    board->copyState = returnstate;
    
  }
  va_end( arglist );
}

static void opening_register_multi(const Board* b, ...) {
  
  va_list arglist;
  va_start(arglist, b);
  const char* ucimove;
  while ((ucimove = va_arg(arglist, const char *)) != NULL) {
    opening_register(b, ucimove);
  }
  va_end( arglist );
}


#define OPENING_START()  {Board temp = FEN_TO_BOARD(temp, STARTPOS_FEN);\

#define OPENING_MOVES(...)  {Board temp = FEN_TO_BOARD(temp, STARTPOS_FEN);\
opening_moves_to_board(&temp, __VA_ARGS__, NULL);

#define OPENING_REGISTER(expr, ...) expr; \
opening_register_multi(&temp, __VA_ARGS__, NULL);\
}

#define OPENING_REGISTER_FEN(fen, ...) {Board temp = FEN_TO_BOARD(temp, fen);\
opening_register_multi(&temp, __VA_ARGS__, NULL);\
}

static move opening_pick(const Opening* opening) {
  int i = rand() % opening->count;
  return opening->m[i];
}

void opening_whitebook(void);
void opening_blackbook(void);

void opening_book_init(void) {
  openings = calloc(BOOK_SIZE, sizeof(*openings));
  unsigned int seed = (int) time(NULL);
  srand(seed);
  opening_whitebook();
  opening_blackbook();
}


void opening_book_set_enabled(bool enabled) {
  disabled = !enabled;
}

move opening_book_consult(const Board* board) {
  if (disabled) return 0;
  Opening* opening = opening_book_find(board);
  if (opening == NULL) return 0;
  else return opening_pick(opening);
  return 0;
}

void opening_book_destroy(void) {
  free(openings);
}

void opening_whitebook(void) {
  OPENING_REGISTER(OPENING_START(),
                   "d2d4", "d2d4", "d2d4", "d2d4", "d2d4",
                   "e2e4", "e2e4", "e2e4",
                   "g1f3",
                   "c2c4",
                   "g2g3");
  //d4 Nf6
  OPENING_REGISTER(OPENING_MOVES("d2d4", "g8f6"),
                   "c2c4");
  //d4 d5
  OPENING_REGISTER(OPENING_MOVES("d2d4", "d7d5"),
                   "c2c4");
  //d4 e6
  OPENING_REGISTER(OPENING_MOVES("d2d4", "e7e6"),
                   "c2c4");

  //e4 c5
  OPENING_REGISTER(OPENING_MOVES("e2e4", "c7c5"),
                   "g1f3");
  //e4 e5
  OPENING_REGISTER(OPENING_MOVES("e2e4", "e7e5"),
                   "g1f3");
  //e4 e6
  OPENING_REGISTER(OPENING_MOVES("e2e4", "e7e6"),
                   "d2d4");
  //e4 c6
  OPENING_REGISTER(OPENING_MOVES("e2e4", "c7c6"),
                   "d2d4");
  //e4 d5
  OPENING_REGISTER(OPENING_MOVES("e2e4", "d7d5"),
                   "e4d5", "e4d5", "e4e5");
}


void opening_blackbook(void) {
  //e4
  OPENING_REGISTER(OPENING_MOVES("e2e4"),
                   "c7c5", "c7c5", "e7e5");
  //d4
  OPENING_REGISTER(OPENING_MOVES("d2d4"),
                   "d7d5", "d7d5", "g8f6");
  
  //c4
  OPENING_REGISTER(OPENING_MOVES("c2c4"),
                   "g8f6",  "c7c5", "c7c5", "e7e6", "e7e5", "c7c6");
  //c4 c5 Nc3
  OPENING_REGISTER(OPENING_MOVES("c2c4", "c7c5", "b1c3"),
                   "g7g6" "g7g6", "b8c6");
  //c4 c5 Nf3
  OPENING_REGISTER(OPENING_MOVES("c2c4", "c7c5", "g1f3"),
                   "g7g6" "g7g6", "g8f6");
  //g3
  OPENING_REGISTER(OPENING_MOVES("g2g3"),
                   "d7d5", "d7d5", "g8f6", "g7g6", "e7e5");
  //f4
  OPENING_REGISTER(OPENING_MOVES("f2f4"),
                   "d7d5", "d7d5", "g8f6", "g7g6", "e7e5");
  //b3
  OPENING_REGISTER(OPENING_MOVES("f2f4"),
                   "d7d5", "d7d5", "g8f6", "g7g6", "e7e5");
  //Nc3
  OPENING_REGISTER(OPENING_MOVES("b1c3"),
                   "c7c5", "d7d5", "g8f6", "g7g6", "e7e5");
  //b4
  OPENING_REGISTER(OPENING_MOVES("b2b4"),
                   "c7c5", "d7d5", "g8f6", "g7g6", "e7e6");
  //e3
  OPENING_REGISTER(OPENING_MOVES("b2b4"),
                   "c7c5", "d7d5", "g8f6", "g7g6", "e7e5");
}
