//
//  time_management.c
//  fornax2_core
//
//  Created by Anders on 05/04/2020.
//  Copyright © 2020 Avividi. All rights reserved.
//

#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include "time_management.h"
#include "../genericutils/debugprint.h"
#include "../genericutils/strings.h"
#include "../genericutils/stopwatch.h"
#include "../genericutils/require.h"
#include "../search/searchmodes.h"
#include "../board/board.h"
#include "../board/definitions.h"
#include "../moving/movegen.h"
#include "../moving/makemove.h"
#include "../boardprint.h"
#include "../eval/eval_definitions.h"

static long configured_overhead = TIME_MANAGEMENT_OVERHEAD_DEFAULT;

static long start;
static long allotted_time;
static bool isTimeManageMode = false;

static void *timer(void *arg) {

  long time = *((long*)arg);
  long seconds = time / 1000;
  long nanoseconds = (time % 1000) * 1000 * 1000;

  nanosleep((const struct timespec[]){{seconds, nanoseconds}}, NULL);
  if (isTimeManageMode) {
    debug_print_l1("time management: stopping\n");
    searchmodes_stop();
  }
  free(arg);
  return NULL;
}

static long getDivisorFromGamePhase(const Board* board) {
  long gamephase = board->friendly->game_phase;
  long m = PHASEVALUE_INITIAL / 3;
  long phase1 = m * 2;
  long phase2 = m;
  if (gamephase > phase1) return 60;
  else if (gamephase > phase2) return 50;
  else return 40;
}

static long time_management_calculate(const Board* board, long remainingTime, long increment) {
  
  if ((remainingTime + increment) < 50) return -1;
  else if (remainingTime <= increment) return increment / 2;
  
  long divisor = getDivisorFromGamePhase(board);
  if (increment > 400) divisor -= 20;
  
  long standartime = remainingTime / divisor + increment - configured_overhead - 20;

  long noise = 20 - rand() % 40;
  long noised = standartime + (standartime / 100 * noise);
  return noised;
}

void time_management_parse_uci(const Board* board, char** tokens, int length) {
  
  long wtime = 0;
  long winc = 0;
  long btime = 0;
  long binc = 0;
  
  for (int i = 0; i < length - 1; ++i) {
    char* token = tokens[i];
    char* next = tokens[i + 1];
    if (string_equals("wtime", token)) wtime = string_to_long(next);
    else if (string_equals("winc", token)) winc = string_to_long(next);
    else if (string_equals("btime", token)) btime = string_to_long(next);
    else if (string_equals("binc", token)) binc = string_to_long(next);
  }
  
  long mytime = board->isWhiteMove ? wtime : btime;
  long myinc = board->isWhiteMove ? winc : binc;
  
  time_management_search(board, mytime, myinc);
}

static move getOnlyLegalMove(const Board* board) {
  Board copy = BOARD_COPY(copy, board);
  move moves[256];
  uint8_t moveCount = movegen_gen(&copy, moves);
  
  BoardCopiedState* prevstate = copy.copyState;
  
  move onlylegal = 0;
  move noLegal = true;
  
  for (uint8_t i = 0; i < moveCount; ++i) {
    move m = moves[i];
    
    BoardCopiedState nextstate;
    bool legal = makemove_make(&copy, &nextstate, m);
    if (legal) {
      if (noLegal) {
        noLegal = false;
        onlylegal = m;
      }
      else {
        onlylegal = 0;
      }
    }
    makemove_takeback(&copy, prevstate, m);
  }
  
  return onlylegal;
}


void time_management_search(const Board* board, long remainingTime, long increment) {
  require(!isTimeManageMode, "BUG time manage get: mode was already on.\n");
  
  long time = time_management_calculate(board, remainingTime, increment);
  
  if (time < 0) {
    debug_print_l1("time management: instant mode\n");
    searchmodes_depth(board, 2);
    return;
  }
  
  move onlylegal = getOnlyLegalMove(board);
  if (onlylegal) {
    searchmodes_best = onlylegal;
    boardprint_printbest(onlylegal);
    return;
  }
  
  allotted_time = time;

  debug_print_l1("time management: movetime %ld\n", time);
  
  isTimeManageMode = true;
  start = stopwatch_wall_start_instance();
  
  pthread_t t2;
  long* ttime = malloc(sizeof(*ttime));
  *ttime = time;
  pthread_create(&t2, NULL, timer, (void*) ttime);
  
  searchmodes_infinite(board);
}

void time_management_send_stopped(void) {
  if (isTimeManageMode) {
    isTimeManageMode = false;
    start = -1;
  }
}

bool time_management_check_stop(void) {
  if (!isTimeManageMode) return false;

  long time = stopwatch_wall_stop_instance(start);
  
  bool stop = time > (allotted_time / 4) * 3;
  
  if (stop) {
    debug_print_l1("time management: stopping early after %ld/%ld\n", time, allotted_time);
  }
  
  return stop;
}

void time_management_set_overhead(long overhead) {
  require(overhead <= TIME_MANAGEMENT_OVERHEAD_MAX, "overhead was too big %d vs max %d\n", overhead, TIME_MANAGEMENT_OVERHEAD_MAX);
  require(overhead >= TIME_MANAGEMENT_OVERHEAD_MIN, "overhead was too small %d vs min %d\n", overhead, TIME_MANAGEMENT_OVERHEAD_MIN);
  configured_overhead = overhead;
  debug_print_l1("ok overhead = %lu\n", overhead);
}
