//
//  transpositions.c
//  fornax2_core
//
//  Created by Anders on 19/06/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "transpositions.h"
#include "zobrist.h"
#include "../boardprint.h"
#include "../board/board.h"
#include "../genericutils/debugprint.h"
#include "../genericutils/require.h"

//static const uint32_t size = 1024 * 1024 * 8;
static const int max_age = 12;
static uint32_t maintenance_counter = 0;
static uint32_t maintenance_freq;
static int size = TRANSPOSITIONS_SIZE_PER_MB * TRANSPOSITIONS_SIZE_MB_DEFAULT;

static Bucket* array;
static int putCount = 0;
static int putIndexCollisions = 0;

static int getNaturalIndex(ttkey hash) {
  
  assert(hash % size >= 0);
//    printf("natural index=%lld\n", (hash % size));
  return hash % size;
}

static inline void transpositions_set_age(Bucket* b, uint8_t age) {
  b->age_and_type = (b->age_and_type & 0xC0) | age;
}

static inline void transpositions_increment_age(Bucket* b) {
  uint8_t age = transpositions_age(b);
  if (age < max_age) b->age_and_type += 2;
}

static inline void transpositions_decrement_age(Bucket* b) {
  uint8_t age = transpositions_age(b);
  if (age) b->age_and_type -= 1;
}

static int getPutIndex(int candidate, ttkey hash, uint8_t depth) {
  int end = candidate - 1;
  
  while (candidate != end) {
    Bucket* prev = array + candidate;
    if (transpositions_age(prev) == 0) {
      return candidate;
    }
    else if (prev->hash == hash) {
      return candidate;
    }
    else {
      putIndexCollisions++;
      
      candidate++;
      if (candidate >= size) {
        candidate = 0;
      }
    }
  }
  assert(false);
  return 0;
}

void transpositions_maintenance(void) {
  for (int i = 0; i < size; ++i) {
    Bucket* b = array + i;
     transpositions_decrement_age(b);
  }
}

void transpositions_put(ttkey hash, uint8_t depth, eval score, move m, tttype type) {
//  int nat = getNaturalIndex(hash);
//  Bucket bucket = {hash, depth, score, UINT32_MAX, m, type};
//  array[nat] = bucket;
  
  putCount++;
  maintenance_counter++;
  if (maintenance_counter % (maintenance_freq) == 0) {
    transpositions_maintenance();
  }

  Bucket bucket = {.hash = hash, .move = m, .eval = score, .depth = depth, .age_and_type = type};
  transpositions_set_age(&bucket, max_age);
  int index = getPutIndex(getNaturalIndex(bucket.hash), bucket.hash, bucket.depth);
  array[index] = bucket;
}

static int pullupLine(int index, int pullup) {
  if (pullup == 0) return index;
  Bucket* current = array + index;
  int natural = getNaturalIndex(current->hash);
  
  if (index == natural) {
    return natural;
  }
  
  int offset = index - 1;
  int found = -1;
  
  while(natural < offset && pullup > 0) {
    if (offset == 0) offset = size - 1;
    Bucket* b = array + offset;
    if (transpositions_age(b) == 0) {
      found = offset;
      
      if (pullup == 1 || offset == natural) {
        break;
      }
      else {
        pullup--;
      }
    }
    offset--;
  }
  if (found >= 0) {
    *(array + found) = *current;
    transpositions_set_age(current, 0);
    return found;
  }
  return index;
}

static int getExistingIndex(int candidate, ttkey hash) {
  
  int end = candidate - 1;
  int pullup = 0;
  
  int found = -1;
  
  while (candidate != end) {
    Bucket* prev = array + candidate;
    if (prev->hash == hash) {
      transpositions_increment_age(prev);
      found = candidate;
      if (pullup == 0) break;
      found = pullupLine(found, pullup);
    }
    else if (transpositions_age(prev) == 0) {
      break;
    }
    else {
      pullupLine(candidate, pullup);
    }
    
    candidate++;
    if (candidate == size) {
      candidate = 0;
    }
  }
  return found >= 0 ? found : candidate;
}

Bucket* transpositions_get(ttkey hash) {
//  uint8_t nat = getNaturalIndex(hash);
//  return &array[nat];
  Bucket* b = array + getExistingIndex(getNaturalIndex(hash), hash);
  return b;
}

void transpositions_clear(void) {
  for (int i = 0; i < size; ++i) {
    Bucket empty = {0,0,0,0,0};
    array[i] = empty;
  }
  putIndexCollisions = 0;
  putCount = 0;
  maintenance_counter = 0;
}

void transpositions_status() {
  int entries = 0;
  int old = 0;
  
  int natural = 0;
  
  int largestOffset = 0;
  int averageOffset = 0;
  
  for (int i = 0; i < size; ++i) {
    Bucket* b = array + i;
    if (transpositions_age(b) > 0) {
      entries++;
      
      int nat = getNaturalIndex(b->hash);
      if (nat == i) natural++;
      else {
        int ii = i;
        if (ii < nat) ii = size + i;
        int offset = ii - nat;
        if (offset > largestOffset) largestOffset = offset;
        averageOffset += offset;
      }
    }
  }
  averageOffset = entries > 0 ? averageOffset / entries : 0;
  printf("entries = %d/%d (%d%%)\n", entries, size, entries / (size / 100));
  printf("natural = %d/%d (%d%%)\n", natural, entries, entries > 0 ? (natural / (entries / 100)) : 0);
  printf("old = %d/%d (%d%%)\n", old, entries, entries > 0 ? (old / (entries / 100)) : 0);
  printf("index collisions pr get = %d/%d (%d%%)\n", putIndexCollisions, putCount, putCount > 0 ? (putIndexCollisions / (putCount / 100)) : 0);
  
  printf("average offset = %d, largest = %d\n", averageOffset, largestOffset);
}

void transpositions_print_line(int i) {
  Bucket* b = array + i;
  printf("[ %c %c | i: %6d | nat: %6d | age: %7d | d: %2d | ",
         (b->depth > 0 ? transpositions_age(b) != 0 ? 'O' : 'A' : ' '),
         (getNaturalIndex(b->hash) == i && transpositions_age(b) != 0 ? 'N' : ' '),
         i,
         getNaturalIndex(b->hash),
         transpositions_age(b),
         b->depth);
  printf(" | hash: %20" PRIu64 "] ",b->hash);
  boardprint_move(b->move);
  printf(" eval %d\n", b->eval);
}

void transpositions_print(long rows) {
  if (rows == 0) {
    rows = size;
  }
  for (int i = 0; i < rows; ++i) {
    transpositions_print_line(i);
  }
}

void transpositions_print_bucket(ttkey hash, long offset) {
  int i = getExistingIndex(getNaturalIndex(hash), hash) + (int) offset;
  transpositions_print_line(i > 3 ? i - 4 : size - 4);
  transpositions_print_line(i > 2 ? i - 3 : size - 3);
  transpositions_print_line(i > 1 ? i - 2 : size - 2);
  transpositions_print_line(i > 0 ? i - 1 : size - 1);
  printf("*\n");
  transpositions_print_line(i);
  printf("*\n");
  transpositions_print_line(i < size - 1 ? i + 1 : 0);
  transpositions_print_line(i < size - 2 ? i + 2 : 2);
  transpositions_print_line(i < size - 3 ? i + 3 : 3);
  transpositions_print_line(i < size - 4 ? i + 4 : 4);
}

void transpositions_init(void) {
  maintenance_freq = size / 24;
  array = malloc(sizeof(struct bucket) * size);
  transpositions_clear();
}

void transpositions_destroy(void) {
  free(array);
}

void transpositions_set_hash_size(long MB) {
  require(MB <= TRANSPOSITIONS_SIZE_MB_MAX, "hash size was too big %lu vs max %d\n", MB, TRANSPOSITIONS_SIZE_MB_MAX);
  require(MB >= TRANSPOSITIONS_SIZE_MB_MIN, "hash size was too small %lu vs min %d\n", MB, TRANSPOSITIONS_SIZE_MB_MIN);
  
  transpositions_destroy();
  size = TRANSPOSITIONS_SIZE_PER_MB * (int) MB;
  transpositions_init();
  debug_print_l1("ok Hash size = %lu\n", MB);
}

