//
//  uci.c
//  fornax2_core
//
//  Created by Anders on 24/04/2019.
//  Copyright © 2019 Avividi. All rights reserved.
//

#ifndef GIT_DESC
#define GIT_DESC "dev"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "uci.h"
#include "genericutils/require.h"
#include "genericutils/strings.h"
#include "genericutils/iolog.h"
#include "board/board.h"
#include "board/move.h"
#include "moving/oracle.h"
#include "moving/makemove.h"
#include "moving/movegen.h"
#include "moving/movepicker.h"
#include "search/searchmodes.h"
#include "search/search.h"
#include "search/search_test_suite.h"
#include "eval/eval.h"
#include "eval/eval_debug.h"
#include "eval/eval_test_suite.h"
#include "perft/perft.h"
#include "perft/perft_test_suite.h"
#include "boardprint.h"
#include "fen/fen.h"
#include "tt/transpositions.h"
#include "supplementary/time_management.h"
#include "supplementary/opening_book.h"

static Board* board;
static char** tokens;

void uci_printversion() {
  char* mainVersion = "1-1";
  if (!string_equals(GIT_DESC, "")) {
    printf("%s %s", mainVersion, GIT_DESC);
  }
  else {
    printf("%s", mainVersion);
  }
}


void uci_init(void) {
  board = board_allocate();
  tokens = malloc(sizeof(*tokens) * 512);
}
void uci_destroy(void) {
  board_free(board);
  free(tokens);
}

static void uci_test(char** tokens, int length) {
  if (string_starts_with(tokens[0], "all")) {
    eval_test_suite_standard();
    perftsuite_all();
    searchsuite_standard();
  }
  else if (string_starts_with(tokens[0], "p")) {
    perftsuite_all();
  }
  else if (string_starts_with(tokens[0], "s")) {
    if (length > 1) {
      int number = (int) string_to_long(tokens[1]);
      searchsuite_number(number);
    }
    else {
      searchsuite_standard();
    }
  }
  else if (string_starts_with(tokens[0], "e")) {
    eval_test_suite_standard();
  }
}

static void uci_tt(char** tokens, int length) {
  if (length == 0) {
    iolog_printf("ttkey = %llu\n", board->copyState->hash);
    iolog_flush();
  }
  else if (string_starts_with(tokens[0], "clear")) {
    transpositions_clear();
  }
  else if (string_starts_with(tokens[0], "status")) {
    transpositions_status();
  }
  else if (string_starts_with(tokens[0], "print")) {
    long size = 0;
    if (length > 1) {
      size = string_to_long(tokens[1]);
    }
    transpositions_print(size);
  }
  else if (string_starts_with(tokens[0], "row")) {
    long size = 0;
    if (length > 1) {
      size = string_to_long(tokens[1]);
    }
    transpositions_print_bucket(board->copyState->hash, size);
  }
}

static void uci_playMoves(char** tokens, int length) {
  int moveListLength = length;
  int i = 0;
  
  BoardCopiedState* previousState = board->copyState;
  
  while (moveListLength > 0) {
    moveListLength--;
    move m = move_full_from_uci(board, tokens[i], (int) strlen(tokens[i]));
    i++;
    BoardCopiedState nextstate;
    bool legal = makemove_make(board, &nextstate, m);
    if (legal) {
      *previousState = nextstate;
      board->copyState = previousState;
      movegen_genmaps_hostile_attacks(board);
    }
    else {
      makemove_takeback(board, previousState, m);
      return;
    }
  }
}

static void uci_position(char** tokens, int length) {
  if (length == 0) {
    return;
  }
  if (!strcmp("fen", tokens[0])) {
    int fenLength = 0;
    while (fenLength++ < length && fenLength <= 6) {
      //todo fix for position fen r3kb1r/pp2pppp/1q2bn2/1N1p4/Q2N4/8/P2BBPPP/R3K2R b KQkq -
      if (string_equals(tokens[fenLength], "moves")) break;
    }
    fen_to_board_split(board, tokens + 1, fenLength - 1);
    if (string_starts_with(tokens[fenLength], "moves")) {
      uci_playMoves(tokens + fenLength + 1, length - fenLength - 1);
    }
    
  }
  else if (string_starts_with(tokens[0], "startpos")) {
    fen_to_board(board, STARTPOS_FEN);
    if (length <= 2) return;
    if (string_starts_with(tokens[1], "moves")) {
      uci_playMoves(tokens + 2, length - 2);
    }
  }
}

static void uci_perft(char** tokens, int length) {
  int depth = 1;
  int threads = 1;

  if (length > 0) {
    require(isdigit(*tokens[0]), "'%s' was not a digit\n", tokens[0]);
    depth = (int)string_to_long(tokens[0]);
    if (length > 1) {
      require(isdigit(*tokens[1]), "'%s' was not a digit\n", tokens[0]);
      threads = (int) string_to_long(tokens[0]);
    }
  }
  perft_run(board, depth, threads);
}


static void uci_go(char** tokens, int length) {
  move bookmove = opening_book_consult(board);
  if (bookmove) {
    searchmodes_best = bookmove;
    boardprint_printbest(bookmove);
    return;
  }
  
  if (length == 0) {
    searchmodes_time(board, 1000);
    return;
  }
  else if (!strcmp(tokens[0], "depth")
           && length == 2
           && isdigit(*tokens[1])) {
    searchmodes_depth(board, (int) string_to_long(tokens[1]));
  }
  else if (string_starts_with(tokens[0], "infinite")) {
     searchmodes_infinite(board);
  }
  else if ((string_starts_with(tokens[0], "movetime")
           || string_starts_with(tokens[0], "time"))
           && length == 2
           && isdigit(*tokens[1])) {
     searchmodes_time(board, (int) string_to_long(tokens[1]));
  }
  else {
    time_management_parse_uci(board, tokens, length);
  }
}

static void uci_setoption(char** tokens, int length) {
  if (length < 4 || !string_equals("name", tokens[0])) return;
  
  if (length == 5
      && string_equals("Move", tokens[1])
      && string_equals("Overhead", tokens[2])) {
    long overhead = string_to_long(tokens[4]);
    time_management_set_overhead(overhead);
  }
  else if (length == 4 && string_equals("Hash", tokens[1])) {
    long MB = string_to_long(tokens[3]);
    transpositions_set_hash_size(MB);
  }
  if (length == 4 && string_equals("internal_opening_book_enabled", tokens[1])) {
    bool enabled = string_to_long(tokens[3]);
    opening_book_set_enabled(enabled);
  }
    
}

static void uci_do(char** tokens, int length) {
  move m;
  if (length == 0) {
    m = searchmodes_best;
  }
  else {
    m = move_full_from_uci(board, tokens[0], (int) strlen(tokens[0]));
  }
  BoardCopiedState* previousState = board->copyState;
  BoardCopiedState nextstate;
  bool legal = makemove_make(board, &nextstate, m);
  if (legal) {
    *previousState = nextstate;
    board->copyState = previousState;
    movegen_genmaps_hostile_attacks(board);
    boardprint_board(*board);
  }
  else {
    makemove_takeback(board, previousState, m);
    puts("illegal");
    return;
  }
}

static bool uci_handleToken(char** tokens, int length) {
  if (length == 0) {
    return true;
  }
  
  char* token = tokens[0];
  
  if (string_equals(token, "go")) {
    uci_go(tokens + 1, length - 1);
  }
  else if (string_equals(token, "stop")) {
    searchmodes_stop();
  }
  else if (string_equals(token, "position")) {
    uci_position(tokens + 1, length - 1);
  }
  else if (string_equals(token, "ucinewgame")) {
    transpositions_clear();
  }
  else if (string_equals(token, "uci")) {
    iolog_printf("id name fornax v");
    uci_printversion();
    iolog_printf("\nid author anders fuglseth\n\n");
    iolog_printf("option name Move Overhead type spin default %d min %d max %d\n",
                 TIME_MANAGEMENT_OVERHEAD_DEFAULT, TIME_MANAGEMENT_OVERHEAD_MIN, TIME_MANAGEMENT_OVERHEAD_MAX);
    iolog_printf("option name Hash type spin default %d min %d max %d\n",
                 TRANSPOSITIONS_SIZE_MB_DEFAULT, TRANSPOSITIONS_SIZE_MB_MIN, TRANSPOSITIONS_SIZE_MB_MAX);
    iolog_printf("uciok\n");
    iolog_flush();
  }
  else if (string_equals(token, "isready")) {
    iolog_printf("readyok\n");
    iolog_flush();
  }
  else if (string_equals(token, "quit")) {
    return false;
  }
  else if (string_equals(token, "setoption")) {
    uci_setoption(tokens + 1, length - 1);
  }
  
  //non uci commands
  else if (string_equals(token, "d")) {
    boardprint_board(*board);
  }
  else if (string_equals(token, "info")) {
    search_printinfo();
  }
  else if (string_equals(token, "do")) {
    uci_do(tokens + 1, length - 1);
  }
  else if (string_equals(token, "perft")) {
    uci_perft(tokens + 1, length - 1);
  }
  else if (string_equals(token, "test")) {
    uci_test(tokens + 1, length - 1);
  }
  else if (string_equals(token, "tt")) {
    uci_tt(tokens + 1, length -1);
  }
  else if (string_equals(token, "eval")) {
    eval_evaluate_debug(board);
  }
  else if (string_equals(token, "moveorder")) {
    Board copy = BOARD_COPY(copy, board);
    Movepick picker;
    movepick_start(&picker, &copy, 0);
    move m;
    while ((m = movepick_next(&picker)) != 0) {
      boardprint_debugmove(m);
    }
  }
  else if (string_equals(token, "pf")) {
    fen_to_board_split(board, tokens + 1, length - 1);
  }
  else if (string_equals(token, "flip")) {
    BoardCopiedState cstate;
    Board copy;
    copy.copyState = &cstate;
    board_flip(board, &copy);
    board_copy(&copy, board);
    boardprint_board(*board);
  }
  else {
    uci_handleToken(tokens + 1, length - 1);
  }
  return true;
}

bool uci_handlecommand(char* command) {
  
  int tokenLength = strings_split(command, tokens, " ");
  if (tokenLength <= 0) return true;

  for (int i = 0; i < tokenLength; ++i) {
    strings_trim(tokens[i]);
  }

  int result = uci_handleToken(tokens, tokenLength);
  return result;
}
